from gevent.socket import wait_read, wait_write
import fcntl
import gevent
import logging
import os
import pty
import setproctitle
import signal
import struct
import subprocess
import sys
import termios

import pyte

from aj.util import BroadcastQueue


class Terminal():
    """
    Terminal emulator based on pyte.
    """

    def __init__(self, manager=None, id=None, command=None, autoclose=False, autoclose_retain=5, redirect=None):
        """
        Setup the environment for a new terminal.

        :param manager: TerminalManager object
        :type manager: TerminalManager
        :param id: Id fo the terminal
        :type id: hex
        :param command: Command to run
        :type command: string
        :param autoclose: Parameter to close the terminal after the command
        :type autoclose: bool
        :param autoclose_retain: Time to wait before closing in seconds
        :type autoclose_retain: integer
        :param redirect: Default redirect URL after closing
        :type redirect: string
        """

        self.width = 80
        self.height = 25
        self.id = id
        self.manager = manager
        self.autoclose = autoclose
        self.autoclose_retain = autoclose_retain
        if redirect:
            self.redirect = redirect
        else:
            self.redirect = '/view/terminal'
        self.output = BroadcastQueue()

        env = {}
        env.update(os.environ)
        env['TERM'] = 'linux'
        env['COLUMNS'] = str(self.width)
        env['LINES'] = str(self.height)
        env['LC_ALL'] = 'en_US.UTF8'

        self.command = command
        if not self.command:
            shell = os.environ.get('SHELL', None)
            if not shell:
                for sh in ['zsh', 'bash', 'sh']:
                    try:
                        shell = subprocess.check_output(['which', sh])
                        break
                    except Exception as e:
                        pass
            self.command = shell

        args = ['sh', '-c', self.command]
        exe = 'sh'

        logging.info('Activating new terminal: %s', self.command)

        self.pid, self.fd = pty.fork()
        if self.pid == 0:
            setproctitle.setproctitle('%s terminal session #%i' % (sys.argv[0], os.getpid()))
            os.execvpe(exe, args, env)

        logging.info('Subprocess PID %s', self.pid)

        self.dead = False

        fl = fcntl.fcntl(self.fd, fcntl.F_GETFL)
        fcntl.fcntl(self.fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

        self.stream_in = os.fdopen(self.fd, 'rb', 0)
        self.stream_out = os.fdopen(self.fd, 'wb', 0)
        self.pyte_stream = pyte.Stream()
        self.screen = pyte.DiffScreen(self.width, self.height)
        self.pyte_stream.attach(self.screen)

        self.last_cursor_position = None

        self.reader = gevent.spawn(self.reader_fn)

    def reader_fn(self):
        """
        Infinite reader of the file descriptor associated with the shell.
        """

        while True:
            wait_read(self.fd)
            self.run_single_read()
            self._check()
            if self.dead:
                return

    def run_single_read(self):
        """
        Propagate the new data read in the file descriptor.
        """

        try:
            data = self.stream_in.read()
        except IOError:
            data = ''

        if data:
            try:
                self.pyte_stream.feed(data.decode('utf-8'))
                self.broadcast_update()
            except UnicodeDecodeError:
                pass

    def broadcast_update(self):
        """
        Send the new formated data to the broadcast queue.
        """

        self.output.broadcast(self.format())

    def _check(self):
        """
        Check if a child process of the shell pid is available.
        """

        try:
            pid, status = os.waitpid(self.pid, os.WNOHANG)
        except OSError:
            self.on_died(code=0)
            return
        if pid:
            self.on_died(code=status)

    def on_died(self, code=0):
        """
        Change terminal status to dead, broadcast new status and autoclose it.

        :param code: Status code of the pid
        :type code: integer
        """

        if self.dead:
            return

        logging.info('Terminal %s has died', self.id)
        self.dead = True

        if code:
            self.pyte_stream.feed(u'\n\n * ' + u'Process has exited with status %i' % code)
        else:
            self.pyte_stream.feed(u'\n\n * ' + u'Process has exited successfully')

        self.broadcast_update()

        if self.autoclose:
            gevent.spawn_later(self.autoclose_retain, self.manager.remove, self.id)

        """
        TODO
        if self.callback:
            try:
                self.callback(exitcode=code)
            except TypeError:
                self.callback()
        """

    def has_updates(self):
        """
        Test if the user made some changes.

        :return: Bool
        :rtype: bool
        """

        if self.last_cursor_position != (self.screen.cursor.x, self.screen.cursor.y):
            return True
        return len(self.screen.dirty) > 0

    def format(self, full=False):
        """
        Prepare the terminal content and attributes to send to the frontend.

        :param full: To send the full buffer
        :type full: bool
        :return: Attributes and content of the terminal
        :rtype: dict
        """

        def compress(line):
            return [[tok or 0 for tok in ch] for _, ch in line.items()]

        l = {}
        self.screen.dirty.add(self.screen.cursor.y)
        if self.last_cursor_position:
            self.screen.dirty.add(self.last_cursor_position[1])
        for k in self.screen.dirty:
            if k < len(self.screen.buffer):
                l[k] = compress(self.screen.buffer[k])
        self.screen.dirty.clear()

        if full:
            l = [compress(x) for _, x in self.screen.buffer.items()]

        r = {
            'lines': l,
            'cx': self.screen.cursor.x,
            'cy': self.screen.cursor.y,
            'cursor': not self.screen.cursor.hidden,
            'h': self.screen.lines,
            'w': self.screen.columns,
        }

        self.last_cursor_position = (self.screen.cursor.x, self.screen.cursor.y)
        return r

    def feed(self, data):
        """
        Write input data into the file descriptor.

        :param data: New data from the user
        :type data: string
        """

        wait_write(self.fd)
        self.stream_out.write(data.encode('utf-8'))
        self.stream_out.flush()

    def resize(self, w, h):
        """
        Resize terminal to new width and height.

        :param w: Width
        :type w: integer
        :param h: Height
        :type h: integer
        """

        if self.dead:
            return
        if (h, w) == (self.screen.lines, self.screen.columns):
            return
        if w <= 0 or h <= 0:
            return
        self.width = w
        self.height = h
        logging.debug('Resizing terminal to %sx%s', w, h)
        self.screen.resize(h, w)
        winsize = struct.pack("HHHH", h, w, 0, 0)
        fcntl.ioctl(self.fd, termios.TIOCSWINSZ, winsize)

    def kill(self):
        """
        Kill this terminal through killing the process associated with the pid.
        """

        self.reader.kill(block=False)

        logging.info('Killing subprocess PID %s', self.pid)
        try:
            os.killpg(self.pid, signal.SIGTERM)
        except OSError:
            pass
        try:
            os.kill(self.pid, signal.SIGKILL)
        except OSError:
            pass



"""
Module to generate virtual shell terminals and interact with the server.
"""

from PIL import Image, ImageDraw

from io import BytesIO
import subprocess

from jadi import component
from aj.api.http import url, HttpPlugin, SocketEndpoint
from aj.auth import authorize
from aj.api.endpoint import endpoint, EndpointError

from .manager import TerminalManager


@component(HttpPlugin)
class Handler(HttpPlugin):
    def __init__(self, context):
        self.context = context
        self.mgr = TerminalManager.get(self.context)

    @url('/api/terminal/script')
    @authorize('terminal:scripts')
    @endpoint(api=True)
    def handle_script(self, http_context):
        """
        Run a script on the server and return the output.
        Method POST.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :return: Return codes (output, errors) from shell
        :rtype: dict
        """

        data = http_context.json_body()
        try:
            p = subprocess.Popen(
                ['bash', '-c', data['script']],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                close_fds=True
            )
            o, e = p.communicate(data.get('input', None))
        except subprocess.CalledProcessError as e:
            raise EndpointError(e)
        except OSError as e:
            raise EndpointError(e)
        return {
            'code': p.returncode,
            'output': o,
            'stderr': e,
        }

    @url('/api/terminal/is_dead/(?P<terminal_id>.+)')
    @endpoint(api=True)
    def handle_test_dead(self, http_context, terminal_id=None):
        """
        Test if a terminal is still open.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param terminal_id: Terminal id
        :type terminal_id: hex
        :return: Alive or not
        :rtype: bool
        """

        if terminal_id in self.mgr:
            return self.mgr[terminal_id].dead
        return True

    @url('/api/terminal/list')
    @endpoint(api=True)
    def handle_list(self, http_context):
        """
        Return the list of opened terminals.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :return: List of terminals ids
        :rtype: list of hex
        """

        return self.mgr.list()

    @url('/api/terminal/create')
    @authorize('terminal:open')
    @endpoint(api=True)
    def handle_create(self, http_context):
        """
        Launch a new virtual terminal with given options.
        Method POST.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :return: Id of the new terminal
        :rtype: hex
        """

        options = http_context.json_body()
        return self.mgr.create(**options)

    @url(r'/api/terminal/kill/(?P<terminal_id>.+)')
    @endpoint(api=True)
    def handle_kill(self, http_context, terminal_id=None):
        """
        Kill a specified terminal.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param terminal_id: Id of the terminal
        :type terminal_id: hex
        :return: Url to redirect
        :rtype: string
        """

        if terminal_id in self.mgr:
            redirect = self.mgr[terminal_id].redirect
            self.mgr.kill(terminal_id)
            return redirect
        return '/view/terminal'

    @url(r'/api/terminal/full/(?P<terminal_id>.+)')
    @endpoint(api=True)
    def handle_full(self, http_context, terminal_id=None):
        """
        Launch a full reload of the terminal.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param terminal_id: Id of the terminal
        :type terminal_id: hex
        :return: Terminal content
        :rtype: dict
        """

        if terminal_id in self.mgr:
            return self.mgr[terminal_id].format(full=True)
        return None

    colors = {
        'black': '#073642',
        'white': '#eee8d5',
        'green': '#859900',
        'yellow': '#b58900',
        'red': '#dc322f',
        'magenta': '#d33682',
        'violet': '#6c71c4',
        'blue': '#268bd2',
        'cyan': '#2aa198',
    }

    @url(r'/api/terminal/preview/(?P<terminal_id>.+)')
    @endpoint(page=True)
    def handle_preview(self, http_context, terminal_id):
        """
        Generate a thumbnail as png for a opened terminal.

        :param http_context: HttpContext
        :type http_context: HttpContext
        :param terminal_id: Id of the terminal
        :type terminal_id: hex
        :return: PNG image
        :rtype: png
        """

        terminal = self.mgr[terminal_id]

        img = Image.new('RGB', (terminal.width, terminal.height * 2))
        draw = ImageDraw.Draw(img)
        draw.rectangle(
            [0, 0, terminal.width, terminal.height],
            fill=ImageDraw.ImageColor.getcolor(self.colors['black'], 'RGB')
        )

        for y in range(0, terminal.height):
            for x in range(0, terminal.width):
                fc = terminal.screen.buffer[y][x][1]
                if fc == 'default':
                    fc = 'lightgray'
                if fc in self.colors:
                    fc = self.colors[fc]
                fc = ImageDraw.ImageColor.getcolor(fc, 'RGB')
                bc = terminal.screen.buffer[y][x][2]
                if bc == 'default':
                    bc = 'black'
                if bc in self.colors:
                    bc = self.colors[bc]
                bc = ImageDraw.ImageColor.getcolor(bc, 'RGB')
                ch = terminal.screen.buffer[y][x][0]
                draw.point((x, y * 2 + 1), fill=(fc if ord(ch) > 32 else bc))
                draw.point((x, y * 2), fill=bc)

        sio = BytesIO()
        img.save(sio, 'PNG')

        http_context.add_header('Content-Type', 'image/png')
        http_context.respond_ok()
        return sio.getvalue()


@component(SocketEndpoint)
class Socket(SocketEndpoint):
    plugin = 'terminal'

    def __init__(self, context):
        SocketEndpoint.__init__(self, context)
        self.mgr = TerminalManager.get(self.context)
        self.readers = {}

    def on_message(self, message, *args):
        terminal_id = message['id']
        if terminal_id not in self.mgr:
            self.send({
                'id': terminal_id,
                'type': 'closed',
            })
            return
        if message['action'] == 'subscribe':
            if terminal_id in self.readers:
                self.readers[terminal_id].kill(block=False)
            self.readers[terminal_id] = self.spawn(self.reader, terminal_id)
        if message['action'] == 'input':
            terminal = self.mgr[terminal_id]
            terminal.feed(message['data'])
        if message['action'] == 'resize':
            terminal = self.mgr[terminal_id]
            terminal.resize(message['width'], message['height'])

    def reader(self, terminal_id):
        terminal = self.mgr[terminal_id]
        q = terminal.output.register()
        while True:
            self.send({
                'type': 'data',
                'id': terminal_id,
                'data': q.get(),
            })



import binascii
import os

from jadi import service

from .terminal import Terminal


@service
class TerminalManager():
    """
    Manager class for all opened terminals
    """

    def __init__(self, context):
        self.context = context
        self.terminals = {}

    def __getitem__(self, _id):
        return self.terminals[_id]

    def __contains__(self, _id):
        return _id in self.terminals

    def list(self):
        """
        List all opened terminals.

        :return: List of terminal ids
        :rtype: list of hex
        """

        return [{
            'id': _id,
            'command': self[_id].command,
        } for _id in self.terminals]

    def create(self, **kwargs):
        """
        Open a new virtual terminal and register it.

        :return: Id of the new terminal
        :rtype: hex
        """

        _id = binascii.hexlify(os.urandom(32)).decode('utf-8')
        t = Terminal(self, _id, **kwargs)
        self.terminals[_id] = t
        return _id

    def kill(self, _id):
        """
        Sent a kill signal to Terminal object and remove it from opened terminals.

        :param _id: Id of the terminal
        :type _id: hex
        """

        self.terminals[_id].kill()
        self.remove(_id)

    def remove(self, _id):
        """
        Remove specified terminal from terminals dict.

        :param _id: Id of the terminal
        :type _id: hex
        """

        try:
            self.terminals.pop(_id)
        except KeyError:
            # Terminal already closed by other process
            pass
